from ninja import Router, Query
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import Filters
from hydrothings.utils import entity_or_404, entities_or_404, generate_response_codes
from .schemas import LocationPostBody, LocationPatchBody, LocationListResponse, LocationGetResponse


router = Router(tags=['Locations'])


@router.get(
    '/Locations',
    response=generate_response_codes('list', LocationListResponse),
    by_alias=True,
    url_name='list_location'
)
def list_locations(request: SensorThingsRequest, filters: Filters = Query(...)):
    """
    Get a collection of Location entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/properties" target="_blank">\
      Location Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/relations" target="_blank">\
      Location Relations</a>
    """

    response = request.engine.list(**filters.dict())

    return entities_or_404(response)


@router.get(
    '/Locations({location_id})',
    response=generate_response_codes('get', LocationGetResponse),
    by_alias=True
)
def get_location(request: SensorThingsRequest, location_id: str):
    """
    Get a Location entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/properties" target="_blank">\
      Location Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/relations" target="_blank">\
      Location Relations</a>
    """

    response = request.engine.get(entity_id=location_id)

    return entity_or_404(response, location_id)


@router.post(
    '/Locations',
    response=generate_response_codes('create')
)
def create_location(request: SensorThingsRequest, response: HttpResponse, location: LocationPostBody):
    """
    Create a new Location entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/properties" target="_blank">\
      Location Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/relations" target="_blank">\
      Location Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    location_id = request.engine.create(
        entity_body=location
    )

    response['location'] = request.engine.get_ref(
        entity_id=location_id
    )

    return 201, None


@router.patch(
    '/Locations({location_id})',
    response=generate_response_codes('update')
)
def update_location(request: SensorThingsRequest, location_id: str, location: LocationPatchBody):
    """
    Update an existing Location entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/properties" target="_blank">\
      Location Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/location/relations" target="_blank">\
      Location Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=location_id,
        entity_body=location
    )

    return 204, None


@router.delete(
    '/Locations({location_id})',
    response=generate_response_codes('delete')
)
def delete_location(request: SensorThingsRequest, location_id: str):
    """
    Delete a Location entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=location_id
    )

    return 204, None



# import json
# import pandas as pd
# from django.core.exceptions import ObjectDoesNotExist
# from django.urls.exceptions import Http404
# from pydantic.fields import SHAPE_LIST
# from pydantic import BaseModel
# from geojson_pydantic import Feature
# from sensorthings import models
# from sensorthings.core.engine import SensorThingsAbstractEngine
# from sensorthings.core.schemas import BasePostBody
# from sensorthings.core import components as component_schemas
# from django.conf import settings


class SensorThingsEngine:

    def __init__(self, host: str, scheme: str, path: str, component: str):
        pass


# class SensorThingsEngine(SensorThingsAbstractEngine):
#
#     def __init__(self, host: str, scheme: str, path: str, component: str):
#         self.host = host
#         self.scheme = scheme
#         self.path = path
#         self.model = getattr(models, component)
#
#     def list(self, filters, count, order_by, skip, top, select, expand, result_format):
#         """"""
#
#         query = self.model.objects
#         response = {}
#
#         if filter is not None:
#             query = self.apply_filters(query, filter)
#
#         response_count = self.get_count(query)
#
#         if order_by is not None:
#             query = self.apply_order(query, order_by)
#
#         if select is not None:
#             query = self.apply_select(query, select)
#
#         if expand is not None:
#             query = self.apply_expand(query, expand)
#
#         if count is True:
#             response['count'] = response_count
#
#         queryset = query.values()
#
#         if top is not None or skip != 0:
#             queryset = self.apply_pagination(queryset, top, skip)
#
#         response_df = pd.DataFrame(list(queryset))
#
#         response_df = self.deserialize_json_fields(response_df)
#         response_df = self.build_related_links(response_df, is_collection=True)
#         response_df = self.build_self_links(response_df, is_collection=True)
#
#         response['value'] = response_df.to_dict('records')
#
#         if top is not None and (top + skip) < response_count:
#             response['next_link'] = self.build_next_link(top, skip)
#
#         return response
#
#     def get(self, entity_id: str, component=None):
#         """"""
#
#         query = self.model.objects
#         query = query.filter(pk=entity_id)
#
#         queryset = query.values()
#
#         response_df = pd.DataFrame(list(queryset))
#
#         response_df = self.deserialize_json_fields(response_df)
#         response_df = self.build_related_links(response_df)
#         response_df = self.build_self_links(response_df)
#
#         try:
#             response = response_df.to_dict('records')[0]
#         except IndexError:
#             response = {}
#
#         return response
#
#     def create(self, entity_body, model=None) -> int:
#         """"""
#
#         active_model = model if model is not None else self.model
#         entity_data = {}
#         entity_links = {}
#
#         for field, value in entity_body:
#             nested_entity = entity_body.__fields__.get(field).field_info.extra.get(
#                 'nested_class', ''
#             ).replace('PostBody', '')
#
#             if issubclass(type(value), BasePostBody):
#                 entity_data[f'{field}_id'] = self.create(
#                     entity_body=value,
#                     model=getattr(models, nested_entity)
#                 )
#             elif isinstance(value, list):
#                 entity_links[field] = [
#                     self.create(
#                         entity_body=sub_value,
#                         model=getattr(models, nested_entity)
#                     ) if issubclass(type(sub_value), BasePostBody) else sub_value.id for sub_value in value
#                 ]
#             elif issubclass(type(value), BaseModel):
#                 entity_data[field] = json.dumps(value.dict())
#             elif isinstance(value, dict):
#                 entity_data[field] = json.dumps(value)
#             else:
#                 entity_data[field] = value
#
#         db_entity = active_model(**entity_data)
#         db_entity.save()
#
#         for related_field, related_ids in entity_links.items():
#             if related_ids:
#                 getattr(db_entity, related_field).add(*related_ids)
#
#         return db_entity.id
#
#     def update(self, entity_id, entity_body):
#         """"""
#
#         try:
#             db_entity = self.model.objects.get(pk=entity_id)
#         except ObjectDoesNotExist:
#             raise Http404
#
#         for attr, value in entity_body.dict(exclude_unset=True).items():
#             setattr(db_entity, attr, value)
#
#         db_entity.save()
#
#         return entity_id
#
#     def delete(self, entity_id):
#         """"""
#
#         try:
#             db_entity = self.model.objects.get(pk=entity_id)
#         except ObjectDoesNotExist:
#             raise Http404
#
#         db_entity.delete()
#
#     def apply_filters(self, query, filters):
#         """"""
#
#         # TODO
#
#         return query
#
#     def get_count(self, query):
#         """
#         Returns a count of the objects that would be returned by the given query.
#
#         :param query: The query object to be counted.
#         :return count: A count of the objects in the resulting queryset.
#         """
#
#         return query.count()
#
#     def apply_order(self, query, order_by):
#         """"""
#
#         # TODO
#
#         return query
#
#     def apply_pagination(self, queryset, top, skip):
#         """"""
#
#         if top is not None:
#             queryset = queryset[skip:skip+top]
#         else:
#             queryset = queryset[skip:]
#
#         return queryset
#
#     def apply_select(self, query, select):
#         """"""
#
#         # TODO
#
#         return query
#
#     def apply_expand(self, query, expand):
#         """"""
#
#         # TODO
#
#         return query
#
#     def build_next_link(self, top, skip):
#         """"""
#
#         return f'{self.get_ref()}?$top={top}&$skip={top+skip}'
#
#     def deserialize_json_fields(self, response_df):
#         """"""
#
#         for name, field in getattr(component_schemas, f'{self.model.__name__}Fields').__fields__.items():
#             if field.type_ in [dict, Feature] and not response_df.empty:
#                 response_df[name] = response_df.apply(
#                     lambda row: json.loads(getattr(row, name)) if getattr(row, name) is not None else None,
#                     axis=1
#                 )
#             elif isinstance(field, BaseModel) and not response_df.empty:
#                 response_df[name] = response_df.apply(
#                     lambda row: json.loads(getattr(row, name).dict()) if getattr(row, name) is not None else None,
#                     axis=1
#                 )
#
#         return response_df
#
#     def build_related_links(self, response_df, is_collection=False):
#         """"""
#
#         for name, field in getattr(component_schemas, f'{self.model.__name__}Relations').__fields__.items():
#             if not response_df.empty:
#                 if field.shape == SHAPE_LIST:
#                     related_component = [
#                         component for component
#                         in settings.ST_CAPABILITIES
#                         if component['SINGULAR_NAME'] == field.type_.__name__
#                     ][0]['NAME']
#                 else:
#                     related_component = field.type_.__name__
#
#                 response_df[f'{name}_link'] = response_df.apply(
#                     lambda row: self.get_ref(row.id if is_collection is True else None, related_component),
#                     axis=1
#                 )
#
#         return response_df
#
#     def build_self_links(self, response_df, is_collection=False):
#         """"""
#
#         if not response_df.empty:
#             response_df['self_link'] = response_df.apply(
#                 lambda row: self.get_ref(row.id if is_collection is True else None),
#                 axis=1
#             )
#
#         return response_df



from ninja.renderers import BaseRenderer
from django.http import HttpRequest
from typing import Any


class FrostServerRenderer(BaseRenderer):
    media_type = "application/json"

    def render(self, request: HttpRequest, data: Any = None, *, response_status: int) -> Any:
        return request



from pydantic import HttpUrl, parse_obj_as


def metadata_validator(value, values):
    """
    Validation for sensor metadata field.

    Runs a check on the sensor metadata field comparing it to the encoding_type field if present. This check ensures
    that the metadata value is consistent with the encoding type.

    :param value: The metadata value passed to the request.
    :param values: The set of values included in the body of the request.
    :return: The output value.
    """

    if values.get('encoding_type') == 'application/pdf':
        try:
            if not parse_obj_as(HttpUrl, value).lower().endswith('pdf'):
                raise AssertionError
        except AssertionError:
            raise ValueError('value is not a valid PDF link')

    elif values.get('encoding_type') == 'http://www.opengis.net/doc/IS/SensorML/2.0':
        try:
            if not parse_obj_as(HttpUrl, value).lower().endswith('xml'):
                raise AssertionError
        except AssertionError:
            raise ValueError('value is not a valid SensorML link')

    elif values.get('encoding_type') == 'text/html':
        try:
            if not parse_obj_as(HttpUrl, value).lower().endswith('html'):
                raise AssertionError
        except AssertionError:
            raise ValueError('value is not a valid HTML link')

    return value



from ninja import Router, Query
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import Filters
from hydrothings.utils import entity_or_404, entities_or_404, generate_response_codes
from .schemas import ThingPostBody, ThingPatchBody, ThingListResponse, ThingGetResponse


router = Router(tags=['Things'])


@router.get(
    '/Things',
    response=generate_response_codes('list', ThingListResponse),
    by_alias=True,
    url_name='list_thing',
    exclude_none=True
)
def list_things(request: SensorThingsRequest, filters: Filters = Query(...)):
    """
    Get a collection of Thing entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/properties" target="_blank">\
      Thing Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/relations" target="_blank">\
      Thing Relations</a>
    """

    response = request.engine.list(**filters.dict())

    return entities_or_404(response)


@router.get(
    '/Things({thing_id})',
    response=generate_response_codes('get', ThingGetResponse),
    by_alias=True
)
def get_thing(request: SensorThingsRequest, thing_id: str):
    """
    Get a Thing entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/properties" target="_blank">\
      Thing Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/relations" target="_blank">\
      Thing Relations</a>
    """

    response = request.engine.get(entity_id=thing_id)

    return entity_or_404(response, thing_id)


@router.post(
    '/Things',
    response=generate_response_codes('create')
)
def create_thing(request: SensorThingsRequest, response: HttpResponse, thing: ThingPostBody):
    """
    Create a new Thing entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/properties" target="_blank">\
      Thing Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/relations" target="_blank">\
      Thing Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    thing_id = request.engine.create(
        entity_body=thing
    )

    response['location'] = request.engine.get_ref(
        entity_id=thing_id
    )

    return 201, None


@router.patch(
    '/Things({thing_id})',
    response=generate_response_codes('update')
)
def update_thing(request: SensorThingsRequest, thing_id: str, thing: ThingPatchBody):
    """
    Update an existing Thing entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/properties" target="_blank">\
      Thing Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/thing/relations" target="_blank">\
      Thing Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=thing_id,
        entity_body=thing
    )

    return 204, None


@router.delete(
    '/Things({thing_id})',
    response=generate_response_codes('delete')
)
def delete_thing(request: SensorThingsRequest, thing_id: str):
    """
    Delete a Thing entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=thing_id
    )

    return 204, None



from typing import TYPE_CHECKING, Literal, Union, List
from pydantic import Field, HttpUrl
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, EntityId, \
    NestedEntity, Filters
from hydrothings.extras.iso_types import ISOTime, ISOInterval
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.datastreams.schemas import Datastream
    from hydrothings.components.featuresofinterest.schemas import FeatureOfInterest


observationTypes = Literal[
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_CategoryObservation',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_CountObservation',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Measurement',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Observation',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_TruthObservation'
]


observationComponents = Literal[
    'id', 'phenomenonTime', 'result', 'resultTime', 'resultQuality', 'validTime', 'parameters', 'FeatureOfInterest/id'
]

observationResultFormats = Literal['dataArray']


class UnitOfMeasurement(Schema):
    name: str
    symbol: str
    definition: HttpUrl


class ObservationFields(Schema):
    phenomenon_time: Union[ISOTime, ISOInterval, None] = Field(None, alias='phenomenonTime')
    result: str
    result_time: Union[ISOTime, None] = Field(..., alias='resultTime')
    result_quality: Union[str, None] = Field(None, alias='resultQuality')
    valid_time: Union[ISOInterval, None] = Field(None, alias='validTime')
    parameters: dict = {}


class ObservationRelations(Schema):
    datastream: 'Datastream'
    feature_of_interest: 'FeatureOfInterest'


class Observation(ObservationFields, ObservationRelations):
    pass


class ObservationDataArray(Schema):
    components: List[observationComponents]
    data_array: List[list] = Field(..., alias='dataArray')

    class Config:
        allow_population_by_field_name = True


class ObservationPostBody(BasePostBody, ObservationFields):
    datastream: Union[EntityId, NestedEntity] = Field(
        ..., alias='Datastream', nested_class='DatastreamPostBody'
    )
    feature_of_interest: Union[EntityId, NestedEntity, None] = Field(
        None, alias='FeatureOfInterest', nested_class='FeatureOfInterestPostBody'
    )


class ObservationDataArrayBody(Schema):
    datastream: EntityId = Field(..., alias='Datastream')
    components: List[observationComponents]
    data_array: List[list] = Field(..., alias='dataArray')


@allow_partial
class ObservationPatchBody(BasePatchBody, ObservationFields):
    datastream: EntityId = Field(..., alias='Datastream')
    feature_of_interest: EntityId = Field(..., alias='FeatureOfInterest')


class ObservationGetResponse(ObservationFields, BaseGetResponse):
    datastream_link: HttpUrl = Field(..., alias='Datastream@iot.navigationLink')
    feature_of_interest_link: HttpUrl = Field(..., alias='FeatureOfInterest@iot.navigationLink')

    class Config:
        allow_population_by_field_name = True


class ObservationListResponse(BaseListResponse):
    value: List[ObservationGetResponse]

    class Config:
        allow_population_by_field_name = True


class ObservationDataArrayResponse(BaseListResponse):
    value: List[ObservationDataArray]

    class Config:
        allow_population_by_field_name = True


class ObservationFilters(Filters):
    result_format: Union[observationResultFormats, None] = Field(None, alias='$resultFormat')



import hydrothings
import hydrothings.components as component_schemas
from django.utils.deprecation import MiddlewareMixin
from django.urls import resolve
from django.urls.exceptions import Http404
from django.http import HttpRequest
from hydrothings.utils import lookup_component
from hydrothings import settings


class SensorThingsMiddleware(MiddlewareMixin):

    @staticmethod
    def process_request(request: HttpRequest) -> None:
        """
        Middleware for resolving nested components in SensorThings URLs.

        The SensorThings specification requires the server to handle nested resource paths and addresses to table
        properties (including values and reference links for the properties). This middleware checks the request URLs
        for these special cases, extracts any extra parameters from these URLs and attaches them to the request, and
        links the request to the correct view function by updating the request's path_info attribute.

        :param request: Django HttpRequest object.
        :return: None
        """

        try:
            resolved_path = resolve(request.path_info)
            if resolved_path.url_name != 'st_complex_handler':
                # Path is not part of the SensorThings app. Proceed normally.
                return None
        except StopIteration:
            # Path is not part of the SensorThings app. Proceed normally.
            return None

        if request.method != 'GET':
            # Nested components are only allowed on GET requests.
            raise Http404

        route_length = len(resolved_path.route.split('/'))
        path_components = request.path_info.split('/')[route_length:]
        path_prefix = '/'.join(request.path_info.split('/')[:route_length])
        component = None
        primary_component = None
        previous_component = None
        endpoint = None

        for i, raw_component in enumerate(path_components):
            path_info = f'{path_prefix}/{raw_component}'
            field_name = None

            try:
                resolved_path = resolve(path_info)
                if resolved_path.url_name is None:
                    # Sub-path may contain field names, or implicit links to related entities.
                    raise Http404
                if resolved_path.url_name.startswith('list'):
                    # This sub-path represents a collection of entities.
                    component = lookup_component(
                        input_value=resolved_path.url_name.replace('list_', ''),
                        input_type='snake_singular',
                        output_type='camel_singular'
                    )
                    field_name = lookup_component(
                        input_value=component,
                        input_type='camel_singular',
                        output_type='snake_plural'
                    )
                    primary_component = component
                    endpoint = f'{path_prefix}/{raw_component}'
                elif resolved_path.url_name.startswith('get'):
                    # This sub-path explicitly represents a single entity.
                    component = lookup_component(
                        input_value=resolved_path.url_name.replace('get_', ''),
                        input_type='snake_singular',
                        output_type='camel_singular'
                    )
                    field_name = lookup_component(
                        input_value=component,
                        input_type='camel_singular',
                        output_type='snake_singular'
                    )
                    primary_component = component
                    endpoint = f'{path_prefix}/{raw_component}'
            except Http404:
                try:
                    # This sub-path may be an implicit relation and needs to be converted to an explicit path.
                    component_plural = lookup_component(
                        input_value=raw_component,
                        input_type='camel_singular',
                        output_type='camel_plural'
                    )
                    field_name = lookup_component(
                        input_value=raw_component,
                        input_type='camel_singular',
                        output_type='snake_singular'
                    )
                    primary_component = raw_component
                    endpoint = f'{path_prefix}/{component_plural}(1)'  # TODO Need to lookup the associated id.
                except StopIteration:
                    # This sub-path may be a field name, $value, or $ref.
                    component = raw_component
                    field_name = lookup_component(
                        input_value=raw_component,
                        input_type='camel_singular',
                        output_type='snake_singular'
                    )

            if previous_component in [c['SINGULAR_NAME'] for c in settings.ST_CAPABILITIES]:
                # Check that this component is a valid child of the previous part of the path.
                if field_name not in getattr(component_schemas, previous_component).__fields__:
                    raise Http404
            elif previous_component in ['$value', '$ref']:
                # $value/$ref must be the last components of a path.
                raise Http404
            elif component == '$value':
                # The previous component must be a non-relational field.
                pass  # TODO Need to verify that the previous component is not a related field.
            elif component == '$ref':
                # The previous component must be a non-relational field.
                pass  # TODO Need to verify that the previous component is not a related field.
            else:
                pass  # TODO Need to figure out any other cases that need to be handled here.

            previous_component = component

        request.component_path = '/'.join(path_components)
        if endpoint:
            request.path_info = endpoint
        if primary_component in [c['SINGULAR_NAME'] for c in settings.ST_CAPABILITIES]:
            request.component = primary_component

    @staticmethod
    def process_view(request, view_func, view_args, view_kwargs) -> None:
        """
        Middleware for initializing a datastore engine for the request.

        This middleware generates a SensorThings engine object and attaches it to the request instance. The engine
        should include a connection to the associated database and methods for performing basic CRUD operations on that
        database and information model.

        :param request: Django HttpRequest object.
        :param view_func: The view function associated with this request.
        :param view_args: The arguments that will be passed to the view function.
        :param view_kwargs: The keyword arguments that will be passed to the view function.
        :return: None
        """

        if hasattr(getattr(view_func, '__self__', None), 'api'):
            st_api = view_func.__self__.api
            if isinstance(st_api, hydrothings.SensorThingsAPI):
                if not hasattr(request, 'component'):
                    request.component = lookup_component(
                        input_value=request.path_info.split('/')[-1].split('(')[0],
                        input_type='camel_plural',
                        output_type='camel_singular'
                    )
                if not hasattr(request, 'component_path'):
                    request.component_path = request.path_info.split('/')[-1]
                request.engine = st_api.engine(
                    host=request.get_host(),
                    scheme=request.scheme,
                    path=request.path_info,
                    version=st_api.version,
                    component=request.component,
                    component_path=request.component_path
                )



from typing import TYPE_CHECKING, Union, List
from pydantic import Field, HttpUrl
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, EntityId, NestedEntity
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.locations.schemas import Location
    from hydrothings.components.historicallocations.schemas import HistoricalLocation
    from hydrothings.components.datastreams.schemas import Datastream


class ThingFields(Schema):
    name: str
    description: str
    properties: Union[dict, None] = None

    class Config:
        allow_population_by_field_name = True


class ThingRelations(Schema):
    locations: List['Location'] = []
    historical_locations: List['HistoricalLocation'] = []
    datastreams: List['Datastream'] = []


class Thing(ThingFields, ThingRelations):
    pass


class ThingPostBody(BasePostBody, ThingFields):
    locations: List[Union[EntityId, NestedEntity]] = Field(
        [], alias='Locations', nested_class='LocationPostBody'
    )
    historical_locations: List[NestedEntity] = Field(
        [], alias='HistoricalLocations', nested_class='HistoricalLocationPostBody'
    )
    datastreams: List[NestedEntity] = Field(
        [], alias='Datastreams', nested_class='DatastreamPostBody'
    )


@allow_partial
class ThingPatchBody(BasePatchBody, ThingFields):
    locations: List[EntityId] = Field([], alias='Locations')


class ThingGetResponse(ThingFields, BaseGetResponse):
    locations_link: HttpUrl = Field(..., alias='Locations@iot.navigationLink')
    historical_locations_link: HttpUrl = Field(..., alias='HistoricalLocations@iot.navigationLink')
    datastreams_link: HttpUrl = Field(..., alias='Datastreams@iot.navigationLink')


class ThingListResponse(BaseListResponse):
    value: List[ThingGetResponse]


class ThingGetResponseODM(ThingGetResponse):
    properties: str



from ninja import NinjaAPI, Schema, Router
from copy import deepcopy
from django.urls import re_path
from pydantic import BaseModel
from typing import Union, Literal, Type, NewType, List
from hydrothings.backends.sensorthings.engine import SensorThingsEngine
from hydrothings.backends.odm2.engine import SensorThingsEngineODM2
from hydrothings.backends.frostserver.engine import FrostServerEngine
from hydrothings.engine import SensorThingsAbstractEngine
from hydrothings.components.root.views import router as root_router
from hydrothings.components.datastreams.views import router as datastreams_router
from hydrothings.components.featuresofinterest.views import router as featuresofinterest_router
from hydrothings.components.historicallocations.views import router as historicallocations_router
from hydrothings.components.locations.views import router as locations_router
from hydrothings.components.observations.views import router as observations_router
from hydrothings.components.observedproperties.views import router as observedproperties_router
from hydrothings.components.sensors.views import router as sensors_router
from hydrothings.components.things.views import router as things_router
from hydrothings.utils import generate_response_codes


class SensorThingsAPI(NinjaAPI):

    def __init__(
            self,
            backend: Literal['sensorthings', 'odm2', 'frostserver', None] = None,
            engine: Union[Type[NewType('SensorThingsEngine', SensorThingsAbstractEngine)], None] = None,
            components: Union[List['SensorThingsComponent'], None] = None,
            endpoints: Union[List['SensorThingsEndpoint'], None] = None,
            id_qualifier: str = '',
            **kwargs
    ):

        if not kwargs.get('version'):
            kwargs['version'] = '1.1'

        if kwargs.get('version') not in ['1.0', '1.1']:
            raise ValueError('Unsupported SensorThings version. Supported versions are: 1.0, 1.1')

        if backend not in ['sensorthings', 'odm2', 'frostserver', None]:
            raise ValueError(
                'Unsupported SensorThings backend. Supported backends are: "sensorthings", "odm2", "frostserver"'
            )

        if not backend and not isinstance(engine, type(SensorThingsAbstractEngine)):
            raise ValueError('No backend was specified, and no engine class was defined.')

        super().__init__(**kwargs)

        self.components = components if components is not None else []
        self.endpoints = endpoints if endpoints is not None else []
        self.id_qualifier = id_qualifier

        if backend == 'sensorthings':
            self.engine = SensorThingsEngine
        elif backend == 'odm2':
            self.engine = SensorThingsEngineODM2
        elif backend == 'frostserver':
            self.engine = FrostServerEngine
        else:
            self.engine = engine

        self.add_router('', deepcopy(root_router))
        self.add_router('', self._build_sensorthings_router('datastream', datastreams_router))
        self.add_router('', self._build_sensorthings_router('feature_of_interest', featuresofinterest_router))
        self.add_router('', self._build_sensorthings_router('historical_location', historicallocations_router))
        self.add_router('', self._build_sensorthings_router('location', locations_router))
        self.add_router('', self._build_sensorthings_router('observation', observations_router))
        self.add_router('', self._build_sensorthings_router('observed_property', observedproperties_router))
        self.add_router('', self._build_sensorthings_router('sensor', sensors_router))
        self.add_router('', self._build_sensorthings_router('thing', things_router))

    def _get_urls(self):

        urls = super()._get_urls()
        urls.append(re_path(r'^.*', lambda request: None, name='st_complex_handler'))

        return urls

    def _build_sensorthings_router(self, component, router):
        """"""

        component_settings = next(iter([
            c for c in self.components if c.name == component
        ]), None)

        # endpoint_settings = {
        #     endpoint.name.split('_')[0]: endpoint
        #     for endpoint in self.endpoints
        #     if '_'.join(endpoint.name.split('_')[1:]) == component
        # } if self.endpoints else {}

        st_router = Router(tags=router.tags)

        for path, path_operation in router.path_operations.items():
            for operation in path_operation.operations:
                view_func = deepcopy(operation.view_func)
                response_schema = getattr(operation.response_models.get(200), '__annotations__', {}).get('response')
                operation_method = operation.view_func.__name__.split('_')[0]

                if getattr(component_settings, 'component_schema', None) is not None:
                    if view_func.__annotations__.get(component):
                        for field, schema in component_settings.component_schema.__fields__.items():
                            view_func.__annotations__[component].__fields__[field] = schema

                    if response_schema:
                        if operation_method == 'list':
                            for field, schema in component_settings.component_schema.__fields__.items():
                                response_schema.__fields__['value'].type_.__fields__[field] = schema

                        elif operation_method == 'get':
                            for field, schema in component_settings.component_schema.__fields__.items():
                                response_schema.__fields__[field] = schema

                getattr(st_router, operation.methods[0].lower())(
                    path.replace('(', f'({self.id_qualifier}').replace(')', f'{self.id_qualifier})'),
                    response=generate_response_codes(operation_method, response_schema),
                    by_alias=True
                )(view_func)

        return st_router


class SensorThingsEndpoint(BaseModel):
    name: str
    enabled: bool = True


class SensorThingsComponent(BaseModel):
    name: str
    component_schema: Union[Type[Schema], None] = None



from uuid import UUID
from pydantic import Field, Extra, HttpUrl, validator
from typing import Union
from ninja import Schema
from hydrothings.validators import nested_entities_check, whitespace_to_none


class EntityId(Schema):
    id: UUID = Field(..., alias='@iot.id')

    class Config:
        allow_population_by_field_name = True


class NestedEntity(Schema):

    class Config:
        extra = Extra.allow


class EntityNotFound(Schema):
    message: str


class BasePostBody(Schema):

    _nested_entity_validator = validator(
        '*',
        allow_reuse=True,
        check_fields=False
    )(nested_entities_check)

    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True


class BasePatchBody(Schema):

    _whitespace_validator = validator(
        '*',
        allow_reuse=True,
        check_fields=False,
        pre=True
    )(whitespace_to_none)

    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True


class BaseListResponse(Schema):
    count: Union[int, None] = Field(None, alias='@iot.count')
    value: list = []
    next_link: Union[HttpUrl, None] = Field(None, alias='@iot.nextLink')

    class Config:
        allow_population_by_field_name = True


class BaseGetResponse(Schema):
    id: UUID = Field(..., alias='@iot.id')
    self_link: HttpUrl = Field(..., alias='@iot.selfLink')

    class Config:
        allow_population_by_field_name = True


class Filters(Schema):
    filters: str = Field(None, alias='$filter')
    count: bool = Field(None, alias='$count')
    order_by: str = Field(None, alias='$orderby')
    skip: int = Field(0, alias='$skip')
    top: int = Field(None, alias='$top')
    select: str = Field(None, alias='$select')
    expand: str = Field(None, alias='$expand')



from abc import ABCMeta, abstractmethod
from pydantic.fields import SHAPE_LIST
from typing import Union
from django.http import HttpRequest
from hydrothings import components as component_schemas
from hydrothings import settings


class SensorThingsAbstractEngine(metaclass=ABCMeta):

    scheme: str
    host: str
    path: str
    version: str
    component: str
    component_path: str

    def get_ref(self, entity_id: Union[str, None] = None, related_component: Union[str, None] = None) -> str:
        """
        Builds a reference URL for a given entity.

        :param entity_id: The id of the entity.
        :param related_component: The related component to be appended to the ref URL.
        :return: The entity's reference URL.
        """

        ref_url = f'{self.scheme}://{self.host}{self.path}'

        if entity_id is not None:
            ref_url = f'{ref_url}({entity_id})'

        if related_component is not None:
            ref_url = f'{ref_url}/{related_component}'

        return ref_url

    def get_related_components(self):
        """"""

        return {
            name: [
                component for component
                in settings.ST_CAPABILITIES
                if component['SINGULAR_NAME'] == field.type_.__name__
            ][0]['NAME'] if field.shape == SHAPE_LIST
            else field.type_.__name__
            for name, field in getattr(component_schemas, f'{self.component}Relations').__fields__.items()
        }

    def build_related_links(self, response, is_collection=False):
        """"""

        return [
            dict(
                entity,
                **{
                    f'{name}_link': self.get_ref(
                        entity['id'] if is_collection is True else None,
                        related_component
                    ) for name, related_component in self.get_related_components().items()
                }
            ) for entity in response
        ]

    def build_self_links(self, response, is_collection=False):
        """"""

        return [
            dict(
                entity,
                **{
                    'self_link': self.get_ref(entity['id'] if is_collection is True else None)
                }
            ) for entity in response
        ]

    @abstractmethod
    def list(
            self,
            filters,
            count,
            order_by,
            skip,
            top,
            select,
            expand
    ) -> dict:
        """"""

        pass

    @abstractmethod
    def get(
            self,
            entity_id
    ) -> dict:
        """"""

        pass

    @abstractmethod
    def create(
            self,
            entity_body,
    ) -> str:
        """"""

        pass

    @abstractmethod
    def update(
            self,
            entity_id,
            entity_body
    ) -> str:
        """"""

        pass

    @abstractmethod
    def delete(
            self,
            entity_id
    ) -> None:
        """"""

        pass


class SensorThingsRequest(HttpRequest):
    engine: SensorThingsAbstractEngine



from typing import List
from pydantic import Field
from ninja import Schema


class ServerSettings(Schema):
    conformance: List[str]


class ServerCapabilities(Schema):
    name: str = None
    url: str = None


class ServerRootResponse(Schema):
    server_settings: ServerSettings = Field(None, alias='serverSettings')
    server_capabilities: List[ServerCapabilities] = Field(None, alias='value')

    class Config:
        allow_population_by_field_name = True



import requests
from hydrothings.engine import SensorThingsAbstractEngine
from hydrothings.settings import FROST_BASE_URL


class FrostServerEngine(SensorThingsAbstractEngine):

    def __init__(self, host: str, scheme: str, path: str, version: str, component: str, component_path: str):
        self.host = host
        self.scheme = scheme
        self.path = path
        self.version = version
        self.component = component
        self.component_path = component_path
        self.frost_url = '/'.join((FROST_BASE_URL, f'v{self.version}', self.component_path,))

    def list(
            self,
            filters,
            count,
            order_by,
            skip,
            top,
            select,
            expand,
            result_format
    ) -> requests.Response:

        response = requests.get(self.frost_url)

        return response

    def get(
            self,
            entity_id,
            component=None
    ) -> requests.Response:

        response = requests.get(self.frost_url)

        return response

    def create(
            self,
            entity_body,
            component=None
    ) -> str:
        # print('HELLO!!!')
        # response = requests.post(self.frost_url, data=entity_body)
        # print(response)

        return '0'

    def update(
            self,
            entity_id,
            entity_body
    ) -> str:
        return '0'

    def delete(
            self,
            entity_id
    ) -> None:
        return None



from hydrothings.engine import SensorThingsAbstractEngine


class SensorThingsEngineODM2(SensorThingsAbstractEngine):

    def __init__(self, host: str, scheme: str, path: str, component: str):
        self.host = host
        self.scheme = scheme
        self.path = path
        self.component = component

    def list(
            self,
            filters,
            count,
            order_by,
            skip,
            top,
            select,
            expand,
            result_format
    ) -> dict:
        return {}

    def get(
            self,
            entity_id,
            component=None
    ) -> dict:
        return {}

    def create(
            self,
            entity_body,
            component=None
    ) -> str:
        return '0'

    def update(
            self,
            entity_id,
            entity_body
    ) -> str:
        return '0'

    def delete(
            self,
            entity_id
    ) -> None:
        return None



from django.urls import path
from hydrothings import SensorThingsAPI


hydrothings_api_1_1 = SensorThingsAPI(
    version='1.1',
    backend='frostserver'
)

hydrothings_api_1_0 = SensorThingsAPI(
    version='1.0',
    backend='frostserver'
)

urlpatterns = [
    path('v1.0/', hydrothings_api_1_0.urls),
    path('v1.1/', hydrothings_api_1_1.urls)
]



from typing import Union
from hydrothings.schemas import EntityId
from .schemas import ObservationPostBody

fields = [
    ('id', 'id',),
    ('phenomenon_time', 'phenomenonTime',),
    ('result_time', 'resultTime',),
    ('result', 'result',),
    ('result_quality', 'resultQuality',),
    ('valid_time', 'validTime',),
    ('parameters', 'parameters',),
    ('feature_of_interest', 'FeatureOfInterest/id',)
]


def convert_to_data_array(response, select: Union[list, None] = None):
    """"""

    if select:
        selected_fields = [
            field for field in fields if field[0] in select
        ]
    else:
        selected_fields = [
            field for field in fields if field[0] in ['id', 'result_time', 'result']
        ]

    response['value'] = [{
        'components': [
            field[1] for field in selected_fields
        ],
        'data_array': [
            [
                entity[field[0]] for field in selected_fields
            ] for entity in response['value']
        ]
    }]

    return response


def parse_data_array(observation):
    """"""

    observations = []

    for datastream in observation:
        datastream_fields = [
            (field[0], field[1], get_field_index(datastream.components, field[1]),) for field in fields
        ]

        observations.extend([
            ObservationPostBody(
                datastream=datastream.datastream,
                **{
                    datastream_field[0]: entity[datastream_field[2]]
                    if datastream_field[0] != 'feature_of_interest'
                    else EntityId(
                        id=entity[datastream_field[2]]
                    )
                    for datastream_field in datastream_fields if datastream_field[2] is not None
                }
            ) for entity in datastream.data_array
        ])

    return observations


def get_field_index(components, field):
    """"""

    try:
        return components.index(field)
    except ValueError:
        return None



from ninja import Router, Query
from typing import Union, List
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.utils import entity_or_404, entities_or_404, generate_response_codes
from .schemas import ObservationPostBody, ObservationPatchBody, ObservationListResponse, ObservationGetResponse, \
    ObservationFilters, ObservationDataArrayResponse, ObservationDataArrayBody
from .utils import convert_to_data_array, parse_data_array


router = Router(tags=['Observations'])


@router.get(
    '/Observations',
    response=generate_response_codes('list', Union[ObservationListResponse, ObservationDataArrayResponse]),
    by_alias=True,
    url_name='list_observation',
    exclude_none=True
)
def list_observations(request: SensorThingsRequest, filters: ObservationFilters = Query(...)):
    """
    Get a collection of Observation entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a>
    """

    request_filters = filters.dict()
    result_format = request_filters.pop('result_format', None)

    response = request.engine.list(**request_filters)

    if result_format == 'dataArray':
        response = convert_to_data_array(response)

    return entities_or_404(response)


@router.get(
    '/Observations({observation_id})',
    response=generate_response_codes('get', ObservationGetResponse),
    by_alias=True
)
def get_observation(request: SensorThingsRequest, observation_id: str):
    """
    Get an Observation entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a>
    """

    response = request.engine.get(entity_id=observation_id)

    return entity_or_404(response, observation_id)


@router.post(
    '/Observations',
    response=generate_response_codes('create')
)
def create_observation(
        request: SensorThingsRequest,
        response: HttpResponse,
        observation: Union[ObservationPostBody, List[ObservationDataArrayBody]]
):
    """
    Create a new Observation entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    if isinstance(observation, list):
        observations = parse_data_array(observation)

        if hasattr(request.engine, 'bulk_create'):
            observation_ids = request.engine.bulk_create(
                entity_bodies=observations
            )
            response_body = [
                request.engine.get_ref(
                    entity_id=observation_id
                ) for observation_id in observation_ids
            ]
        else:
            response_body = []
            for observation in observations:
                observation_id = request.engine.create(
                    entity_body=observation
                )
                response_body.extend(
                    request.engine.get_ref(
                        entity_id=observation_id
                    )
                )

    else:
        observation_id = request.engine.create(
            entity_body=observation
        )

        response['location'] = request.engine.get_ref(
            entity_id=observation_id
        )

        response_body = None

    return 201, response_body


@router.patch(
    '/Observations({observation_id})',
    response=generate_response_codes('update')
)
def update_observation(request: SensorThingsRequest, observation_id: str, observation: ObservationPatchBody):
    """
    Update an existing Observation entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=observation_id,
        entity_body=observation
    )

    return 204, None


@router.delete(
    '/Observations({observation_id})',
    response=generate_response_codes('delete')
)
def delete_observation(request: SensorThingsRequest, observation_id: str):
    """
    Delete a Observation entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=observation_id
    )

    return 204, None



from datetime import datetime


class ISOTime(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError('string required')

        try:
            datetime.fromisoformat(v)
        except TypeError:
            raise ValueError('invalid ISO time format')

        return cls(v)


class ISOInterval(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError('string required')

        split_v = v.split('/')

        try:
            if len(split_v) != 2 or datetime.fromisoformat(split_v[0]) >= datetime.fromisoformat(split_v[1]):
                raise TypeError
        except TypeError:
            raise ValueError('invalid ISO interval format')

        return cls(v)



from typing import TYPE_CHECKING, Literal, List, Union
from pydantic import Field, HttpUrl
from geojson_pydantic import Feature
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, EntityId, NestedEntity
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.things.schemas import Thing
    from hydrothings.components.historicallocations.schemas import HistoricalLocation


locationEncodingTypes = Literal['application/geo+json']


class LocationFields(Schema):
    name: str
    description: str
    encoding_type: locationEncodingTypes = Field(..., alias='encodingType')
    location: Feature
    properties: dict = {}


class LocationRelations(Schema):
    things: List['Thing'] = []
    historical_locations: List['HistoricalLocation'] = []


class Location(LocationFields, LocationRelations):
    pass


class LocationPostBody(BasePostBody, LocationFields):
    things: List[Union[EntityId, NestedEntity]] = Field(
        [], alias='Things', nested_class='ThingPostBody'
    )
    historical_locations: List[Union[EntityId, NestedEntity]] = Field(
        [], alias='HistoricalLocations', nested_class='HistoricalLocationPostBody'
    )


@allow_partial
class LocationPatchBody(LocationFields, BasePatchBody):
    things: List[EntityId] = Field([], alias='Things')
    historical_locations: List[EntityId] = Field([], alias='HistoricalLocations')


class LocationGetResponse(BaseGetResponse, LocationFields):
    things_link: HttpUrl = Field(..., alias='Things@iot.navigationLink')
    historical_locations_link: HttpUrl = Field(..., alias='HistoricalLocations@iot.navigationLink')


class LocationListResponse(BaseListResponse):
    value: List[LocationGetResponse]
